/****************************************************************************
 * elasticWizard.c
 *
 * Copyright (c) 2000 Stefan Hellkvist, Ericsson Radio Systems AB
 * Email: stefan@hellkvist.org
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *****************************************************************************/

/**
 * elasticWizard.c
 * A wizard implementation using elastic matching
 * More info about this method can be found in my
 * Master's thesis report which at the time of this
 * writing can be found at: http://www.testbed.era.ericsson.se/~erasshe
 * If it's not there anymore, contact me at: peffis@mindless.com
 **/

/* Oct 2000 - ported to Pocket Viewer by Wittawat Yamwong <wittawat@web.de> */
/* 2000-2002 adapted for PVMerlin by Christian Ensel */


/*#define ALLOWDEBUGCODE /*ce enables some debug code; removed for PVMerlin V3.1 for code-size reasons */

#include <stdrom.h>
#include "md.h"
#include "mm.h"

#include "pattern.h"
#include "wizard.h"
#include "preproc.h"
#include "filter.h"
#include "def.h"

#include "define.h"
#include "libc.h"

#include "mod_symb.h"
#include "models.h"

#ifdef ALLOWDEBUGCODE
#include "params.h"  /* for debug-option */
#endif

#define T_DIST i32 /*word ce: using word instead of i32 works, but brings no measurable speed-up */
#define THRESHOLD_MAX (0x7fffffff / 256) /* for word: 0xffff; for i32: (0x7fffffff / 256) */

/*#define MAX_PATH 1024
#define MAX_PATTERNS 40*/
#define ABS( x )    _asm_iabs("\ncmp ax,0\n jge _ge\n neg ax\n _ge:\n",x)
#define MIN( x, y ) _asm_imin("\ncmp ax,bx\n jl _le\n mov ax,bx\n _le:\n", x, y)
#define MAX( x, y ) max(x, y)

static filter fs[] = { antiSlantedFilter, sizeNormFilter, }; /* , /*smoothFilter,*/ /* pointDistanceNormFilter }; */

static Preprocessor FAR *preprocessor;
static BANK currentBank = ALPHA;
static Pattern FAR *samples, FAR *unknown, FAR *currentModel;
static T_DIST FAR * FAR *lookup = NULL;
static int lookup_n = 0, lookup_m = 0;

/* Some local functions */
static T_DIST distance( int sampleIndex, T_DIST Threshold );
static int newLookup( int n, int m );
static void freeLookup();
static int maxWidth( Pattern FAR *p );

/* inline function for slopeDistance -- wy */
static int _asm_iabs(char *, int);
static int _asm_imin(char *, int, int);

/*ce alternative character generation does not work well (good and slow or bad...)
static unsigned char c_alt1;
static T_DIST dist_alt1;
*/


/**
 * setBank
 * Guess what?
 **/
void
setBank( BANK b )
{
    currentBank = b;
}


/**
 * getBank
 * Guess what?
 **/
BANK
getBank()
{
    return currentBank;
}


/**
 * initWizard
 * Prepares the wizard for use
 **/
void
initWizard()
{
    preprocessor = newPreprocessor( fs, sizeof(fs)/sizeof(filter) );
    setBank( ALPHA );
}


/**
 * initModels
 * Initiates models
 **/
int
initModels()
{
    return 0;
}


/**
 * freeModels
 * Guess what
 **/
void
freeModels()
{
}



/**
 * ce:
 * doallpreprocessing
 * code cut out of recoginze() 
 * makes a nice & preprocessed pattern from a raw input pattern 
 * (allocates new memory for the new pattern. the old one is also preserved)
 **/
Pattern FAR *
doallpreprocessing( Pattern FAR *p )
{
    Pattern FAR *np;
    np = apply( preprocessor, p );
    if (np == NULL)
        return NULL;
    if (!calcSlopes( np )) {
        deletePattern(np);
        return NULL;
    }
    calcPreDecideInfo( np );
    return np;
}



static unsigned char current_ascii = 0;
static T_DIST current_mind = THRESHOLD_MAX;

/**
 * recognize
 * The magic! Given a (ce: nice&preprocessed!) pattern it returns the ascii that
 * it thinks is closest
 * (ce: does not delete the pattern np (i.e., does not free the memory))
 **/
unsigned char
recognize( Pattern FAR *np )
{
    current_ascii = 0;              /*ce samples[0].ascii; */
    current_mind = THRESHOLD_MAX;   /*ce distance( 0, 0x7fffffff / 256 ); */
    /*c_alt1 = 0;
    dist_alt1 = THRESHOLD_MAX;*/
    return recognizeMore( np, getBank() );
}


/*ce 
 * same as above, but able to continue search on another model */
unsigned char
recognizeMore( Pattern FAR *np, 
               BANK b )    /* ce all getBank() in this function replaces with b */
{
    T_DIST tmpd;
    int i, nsamples;
 
#ifdef ALLOWDEBUGCODE
    /*ce just for test/debugging purposes */
    int ok_sev = 0;
    int ok_sp = 0;
#endif 

    char candidates[50];
    char ci;
    /**
     * Check for shift-symbol
     */
    /*ce -- should not happen in this function: all patterns are pre-scaled*/
    /*if ( maxWidth( p ) < 6 )
          return '.';*/

   
    /*
     * Do the magic stuff
     */
    samples = models[ b ];
    nsamples = npatterns[ b ];
    unknown = np;

    /*ce some pre-decission parameters */
    /*#define CLOSED_LETTER_RANGE (30*30)   /* if sev_l is smaller than this value, the letter is regarded a 'closed' letter, i.e., the endpoint is close to the startpoint */
    /*#define CLOSED_LETTER_FUZZYNESS (0.1) */
#define MAX_VECTOR_DIFF 33  /* maximum allowed for dx and dy (this is faster to test than the real vector distance) */
#define MAX_STARTPOINT_DIFF 45  /* maximum allowed for dx and dy (this is faster to test than the real vector distance) */
 
    /* some decissions are taken outside of the loop => more code, but faster loop */
    /* we could also pre-sort the models, but then changing the models.h-file becomes more complicated */
    ci=0;
    for ( i = 0; i < nsamples; i++ ) {
        /*ce only compare apparently 'similar' patterns (according to the pre-decission info) */
        /*comparing the boxshape was not very successful ...
        #define MAX_BOXSHAPE_DIFF 39
        int dbs =  samples[i].boxshape - np->boxshape;
        if ( dbs < MAX_BOXSHAPE_DIFF && dbs > -MAX_BOXSHAPE_DIFF ) { */
        /* compare start-end-vector */
        int dx = min( abs(samples[i].sev_x - np->sev_x),   /* for straight writers */ 
                      abs(samples[i].sev_x - (np->sev_x +/*ce-new4 - changed sign*/ (np->sev_y>>2)) ) ); /* for slanted writers */
        int dy = samples[i].sev_y - np->sev_y;
        if (  dx < MAX_VECTOR_DIFF &&
              dy < MAX_VECTOR_DIFF && dy > -MAX_VECTOR_DIFF) {
            /* now compare starting point */
            dx = samples[i].pv[0].x - np->pv[0].x;
            dy = samples[i].pv[0].y - np->pv[0].y;
            if ( dx < MAX_STARTPOINT_DIFF && dx > -MAX_STARTPOINT_DIFF &&
                 dy < MAX_STARTPOINT_DIFF && dy > -MAX_STARTPOINT_DIFF ) {
                candidates[ci]=samples[i].ascii;
                ci++;
                tmpd = distance( i, current_mind ); /* dist_alt1 );*/
                if ( tmpd < current_mind ) {
                    /* dist_alt1 = tmpd + (tmpd>>3); /* next threshold is current result + 12.5% *
                    if ( current_mind < dist_alt1 ) {
                        /* previous best is within 12.5% -> remember it as second best *
                        dist_alt1 = current_mind;
                        c_alt1 = current_ascii;
                    } else
                        c_alt1 = 0; /* forget last one, its too bad *
                    */
                    /* set new best candidate */
                    current_mind = tmpd;
                    current_ascii = samples[i].ascii;
                } /*
                else if ( tmpd < dist_alt1 ) { /* it's not best, but second best *
                    c_alt1 = tmpd;
                    dist_alt1 = samples[i].ascii;
                } */
            } 
#ifdef ALLOWDEBUGCODE
            else
                ok_sev++;
#endif
        } 
#ifdef ALLOWDEBUGCODE
        else { /*ce just for debugging/statistical reasons */
            dx = samples[i].pv[0].x - np->pv[0].x;
            dy = samples[i].pv[0].y - np->pv[0].y;
            if ( dx < MAX_STARTPOINT_DIFF && dx > -MAX_STARTPOINT_DIFF &&
                 dy < MAX_STARTPOINT_DIFF && dy > -MAX_STARTPOINT_DIFF )
                ok_sp++;
        }
#endif
    } 

/* ce original code:
    for ( i = 0; i < nsamples; i++ ) {
        tmpd = distance( i, mind );
        if ( tmpd < mind ) {
            mind = tmpd;
            ascii = samples[i].ascii;
        }
    }
*/


    candidates[ci]=0;
#ifdef ALLOWDEBUGCODE
    if ( PARAM(PARAM_DEBUG) == P_DBG_PREPROC || PARAM(PARAM_DEBUG) == P_DBG_FULL ) { 
        LibPutMsgDlg("->%c\r"
                     "v:%i,%i n:%i\r"
                     "s:%i,%i\r\r"
                     "ok:%i -sev:%i -sp:%i\r"
                     "%s",
                     current_ascii,
                     np->sev_x, np->sev_y, np->nPoints,
                     np->pv[0].x, np->pv[0].y,
                     ci, ok_sev, ok_sp,
                     candidates );
    }
#endif

    /*
     * cleanup and return
     */

    freeLookup();
    return current_ascii;
}

/*ce-new2
unsigned char getInputAlternative()
{
    return c_alt1;
}

void clearInputAlternative()
{
    c_alt1 = 0;
}
ce-new2-end */

/**
 * maxWidth
 * Returns the maximum width of the pattern
 **/
static
int
maxWidth( Pattern FAR *p )
{
    int minx, maxx, miny, maxy, i;
    Point p1 = p->pv[0];
    minx = maxx = p1.x;
    miny = maxy = p1.y;

    for ( i = 1; i < p->nPoints; i++ )
    {
    Point p2 = p->pv[i];
    if (p2.x > maxx)
        maxx = p2.x;
    if (p2.y > maxy)
        maxy = p2.y;
    if (p2.x < minx)
        minx = p2.x;
    if (p2.y < miny)
        miny = p2.y;
    }
    return MAX( maxx - minx, maxy - miny );
}


/**
 * A real function for returning the min of two values
 * a macro would not be good here, since arguments can be calculated twice then
 **/
static T_DIST lmin( T_DIST x, T_DIST y )
{
    return ( x < y ) ? x : y;
}

/**
 * slopeDistance
 * The distance between two angular values
 **/
static
int
slopeDistance( int a1, int a2 )
{
    int delta = ABS( a1 - a2 );
    return MIN( delta, ((int)( C * 2 * PI )) - delta );
}

#define NORM(x,y) ((x)*(x) + (y)*(y))
/**
 * pointDistance
 * The diatance function between point i in the unknown pattern and
 * point j in the current model
 **/
static
int
pointDistance( int i, int j )
{
    Point p1 = unknown->pv[i];
    Point p2 = currentModel->pv[j];
    return NORM(p1.x - p2.x, p1.y - p2.y)
    + slopeDistance( unknown->s[i], currentModel->s[j] );
}

/**
 * freeLookup
 * Frees the lookup table
 **/
static
void
freeLookup()
{
    if (lookup == NULL)
    return;
    if (lookup[0] != NULL)
    free(lookup[0]);
    free( lookup );
    lookup = NULL;
    lookup_m = 0;
    lookup_n = 0;
}


/**
 * newLookup
 * Initializes the lookup-table
 **/
static
int
newLookup( int n, int m )
{
    int i;
    if (n != lookup_n) {
    freeLookup();
    lookup = (T_DIST FAR * FAR *) mm_malloc( (lookup_n = n) * sizeof(T_DIST FAR *));
    if (lookup == NULL)
        return 0;
    lookup[0] = NULL;
    }
    if (m > lookup_m) {
    T_DIST FAR *buf;
    if (lookup[0] != NULL) {
        free(lookup[0]);
        lookup[0] = NULL;
        lookup_m = 0;
    }
    m = m + 5; /* preallocate -- wy */
    if ((buf = (T_DIST FAR *) mm_malloc( m * n * sizeof(T_DIST))) == NULL)
        return 0;
    for (i = 0; i < lookup_n; i++) {
        lookup[i] = buf;
        buf += m;
    }
        lookup_m = m;
    }

}

#if 0
/**
 * The actual Recursive Elastic Matching function
 * See my thesis for some nicely printed equations
 **/
T_DIST
D( int i, int j )
{
    if ( lookup[i][j] != -1 )
    return lookup[i][j];
    if ( j == 1 )
    return lookup[i][j] = pointDistance(i, 1)
        + lmin(D(i-1, 1), D(i-1, 0));
    else
    return lookup[i][j] = pointDistance(i,j)
        + lmin(D(i-1, j), lmin(D(i-1, j-1), D(i-1, j-2)));
}
#endif

/* The Elastic Matching function without recusion.
   It'll stop if the result becomes over Threshold. - wy */
static
T_DIST
D(int N, int M, T_DIST Threshold)
{
    int start,i,j,count;
    T_DIST FAR *tmp;

    tmp = &lookup[0][0];
    *tmp = pointDistance(0,0);
    for (j = 1; j <= M; j++)
        *(++tmp) = pointDistance(0,j) + *tmp;
    start = M - 2*N + 2;
    for (i = 1; i <= N; i++) {
        if (start < 0)
            j = 0;
        else
            j = start;
        count = 0;
        if (j == 0) {
            if ((lookup[i][0] = pointDistance(i,0) + lookup[i-1][0])
                < Threshold)
                count++;
            j++;
        }
        if (j == 1) {
            tmp = &lookup[i-1][0];
            if ((lookup[i][1] = pointDistance(i,1) + lmin(tmp[0],tmp[1]))
                < Threshold)
                count++;
            j++;
        }
        while (j <= M) {
            tmp = &lookup[i-1][j-2];
            if ((lookup[i][j] =
                 pointDistance(i,j) + lmin(tmp[0], lmin(tmp[1], tmp[2])))
                < Threshold)
                count++;
            j++;
        }
        if (count == 0)
            return Threshold;
        start += 2;
    }
    return lookup[N][M];
}

/**
 * distance
 * Calculates the distance from then unknown pattern to sample number
 * sampleIndex using elastic matching
 **/
static
T_DIST
distance( int sampleIndex, T_DIST Threshold )
{
    currentModel = &samples[ sampleIndex ];
    if (!newLookup( unknown->nPoints, currentModel->nPoints ))
    return Threshold;
    return D( unknown->nPoints - 1, currentModel->nPoints - 1,
              Threshold * currentModel->nPoints ) / currentModel->nPoints;
}

